import 'package:flutter/foundation.dart';
import 'package:flutter/material.dart';
import 'package:flutter/rendering.dart';

class MaterialStatefulWidget extends StatefulWidget {
  final bool dragged;
  final bool selected;
  final bool enabled;
  final bool error;
  final VoidCallback? onPressed;
  final VoidCallback? onLongPress;
  final bool drawSplash;
  final ValueWidgetBuilder<MaterialStates> builder;
  final Widget? child;

  const MaterialStatefulWidget({
    Key? key,
    this.dragged = false,
    this.selected = false,
    this.enabled = true,
    this.error = false,
    this.onPressed,
    this.onLongPress,
    this.drawSplash = false,
    required this.builder,
    this.child,
  }) : super(key: key);

  @override
  State<MaterialStatefulWidget> createState() => _MaterialStatefulWidgetState();
}

class _MaterialStatefulWidgetState extends State<MaterialStatefulWidget> {
  final MaterialStates _states = MaterialStates._();

  @override
  void initState() {
    super.initState();
    _updateWidgetStates();
  }

  @override
  void didUpdateWidget(MaterialStatefulWidget oldWidget) {
    super.didUpdateWidget(oldWidget);
    _updateWidgetStates();
  }

  @override
  Widget build(BuildContext context) => InkWell(
      canRequestFocus: widget.enabled,
      onFocusChange: (value) => _updateState(MaterialState.focused, value),
      onHighlightChanged: (value) => _updateState(MaterialState.pressed, value),
      onHover: (value) => _updateState(MaterialState.hovered, value),
      splashFactory: widget.drawSplash ? null : NoSplash.splashFactory,
      highlightColor: widget.drawSplash ? null : Colors.transparent,
      onTap: widget.onPressed,
      onLongPress: widget.onLongPress,
      child: widget.builder(context, _states, widget.child));

  void _updateWidgetStates() {
    _states._update(MaterialState.dragged, widget.dragged);
    _states._update(MaterialState.selected, widget.selected);
    _states._update(MaterialState.disabled, !widget.enabled);
    _states._update(MaterialState.error, widget.error);
  }

  void _updateState(MaterialState state, bool value) => setState(() => _states._update(state, value));
}

@immutable
class MaterialStates {
  final Set<MaterialState> _states = <MaterialState>{};

  MaterialStates._();

  bool get hovered => _states.contains(MaterialState.hovered);

  bool get focused => _states.contains(MaterialState.focused);

  bool get pressed => _states.contains(MaterialState.pressed);

  bool get dragged => _states.contains(MaterialState.dragged);

  bool get selected => _states.contains(MaterialState.selected);

  bool get disabled => _states.contains(MaterialState.disabled);

  bool get error => _states.contains(MaterialState.error);

  bool _update(MaterialState state, bool value) => value ? _states.add(state) : _states.remove(state);

  @override
  bool operator ==(Object other) => other is MaterialStates && setEquals(other._states, _states);

  @override
  int get hashCode => hashValues(runtimeType, hashList(_states));

  @override
  String toString() => 'MaterialStates[${_states.join(', ')}]';
}
